require("RouteManager.nut");

class TestBuildRoadVehicle extends MoneyTest {
    d = null;
	e = null;
	v = null;
	
	function DoAction() {
	    v = AIVehicle.BuildVehicle(d, e);
		if (!AIVehicle.IsValidVehicle(v)) {
		    return false;
		}
		return true;
	}
	
	function GetPrice() {
	    return AIEngine.GetPrice(e);
	}
	
	function TryBuild(depot, engine) {
	    d = depot;
		e = engine;
		if (DoMoneyTest()) {
		    return v;
		}
		return null;
	}
}

class TestRefitRoadVehicle extends MoneyTest {
    v = null;
	c = null;
	
	function DoAction() {
	    if (AIExecMode() && AIVehicle.RefitVehicle(v, c)) {
		    return true;
		}
		AIVehicle.SellVehicle(v);
		return false;
	}
	
	function GetPrice() {
	    local cost = AIAccounting();
		AITestMode() && AIVehicle.RefitVehicle(v, c);
		return cost.GetCosts();
	}
	
	function TryRefit(vehicle, cargoId) {
	    v = vehicle;
		c = cargoId;
		return DoMoneyTest();
	}
}

class Route extends RouteManager {
    MIN_VEHICLE_START_COUNT = 5;
    MAX_VEHICLE_COUNT = 25;
    START_VEHICLE_COUNT = 10;

    m_cityFrom = null;
    m_cityTo = null;
    m_stationFrom = null;
    m_stationTo = null;
    m_depotTile = null;
	m_bridgeTiles = null;
    m_cargoClass = null;

    m_engine = null;
    m_lastVehicleAdded = null;
    m_lastVehicleRemoved = null;

    m_sentToDepotList = null;
    m_renewList = null;

    m_activeRoute = null;
    m_expandedFromCount = null;
    m_expandedToCount = null;


    constructor(cityFrom, cityTo, stationFrom, stationTo, depotTile, bridgeTiles, cargoClass, isLoaded = 0) {
        m_cityFrom = cityFrom;
        m_cityTo = cityTo;
        m_stationFrom = stationFrom;
        m_stationTo = stationTo;
        m_depotTile = depotTile;
		m_bridgeTiles = bridgeTiles;
        m_cargoClass = cargoClass;

        m_engine = GetTruckEngine(cargoClass);

        m_lastVehicleAdded = 0;
        m_lastVehicleRemoved = AIDate.GetCurrentDate();;

        m_sentToDepotList = AIList();
        m_renewList = AIList();

        m_activeRoute = true;
        m_expandedFromCount = 0;
        m_expandedToCount = 0;

        if(!isLoaded) {
            addVehiclesToNewRoute(cargoClass);
        }
    }

	function updateBridges();
    function updateEngine();
    function addVehicle(return_vehicle);
    function addVehiclesToNewRoute(cargoClass);
    function GetEngineList(cargoClass);
    function GetTruckEngine(cargoClass);

    function GetEngineList(cargoClass) {
        local engineList = AIEngineList(AIVehicle.VT_ROAD);
		engineList.Valuate(AIEngine.IsBuildable);
		engineList.KeepValue(1);
		
        engineList.Valuate(AIEngine.GetRoadType);
        engineList.KeepValue(AIRoad.ROADTYPE_ROAD);

	    local stationType = cargoClass == AICargo.CC_PASSENGERS ? AIStation.STATION_BUS_STOP : AIStation.STATION_TRUCK_STOP;
		
	    local station1Tiles = AITileList_StationType(AIStation.GetStationID(m_stationFrom), stationType);
	    local articulated_viable = false;
	    for (local tile = station1Tiles.Begin(); station1Tiles.HasNext(); tile = station1Tiles.Next()) {
    	    if (AIRoad.IsDriveThroughRoadStationTile(tile)) {
// 			    AILog.Info(AIBaseStation.GetName(AIStation.GetStationID(m_stationFrom)) + " is articulated viable!")
		        articulated_viable = true;
			    break;
		    }
	    }
		
	    if (articulated_viable) {
	        local station2Tiles = AITileList_StationType(AIStation.GetStationID(m_stationTo), stationType);
	    	articulated_viable = false;
	   		for (local tile = station2Tiles.Begin(); station2Tiles.HasNext(); tile = station2Tiles.Next()) {
    		    if (AIRoad.IsDriveThroughRoadStationTile(tile)) {
//			        AILog.Info(AIBaseStation.GetName(AIStation.GetStationID(m_stationTo)) + " is articulated viable!")
		            articulated_viable = true;
		    	    break;
	    	    }
	   		}
    	}
		
	    if (!articulated_viable) {
	        engineList.Valuate(AIEngine.IsArticulated);
	        engineList.KeepValue(0);
	    }

        engineList.Valuate(AIEngine.CanRefitCargo, Utils.getCargoId(cargoClass));
        engineList.KeepValue(1);

        return engineList;
    }

    function GetTruckEngine(cargoClass) {
        local engineList = GetEngineList(cargoClass);

        engineList.Valuate(AIEngine.GetMaxSpeed);
        engineList.Sort(AIList.SORT_BY_VALUE, true);
		engineList.KeepBottom(1);
		if (engineList.Count() == 0) return m_engine;

        return engineList.Begin();
    }

    function updateEngine() {
	    if (!m_activeRoute) return;

        m_engine = GetTruckEngine(m_cargoClass);
    }
	
	function updateBridges() {
	    if (!m_activeRoute) return;
		
		AIRoad.SetCurrentRoadType(AIRoad.ROADTYPE_ROAD);
	    for (local i = m_bridgeTiles.Begin(); m_bridgeTiles.HasNext(); i = m_bridgeTiles.Next()) {
		    local north_tile = i;
			local south_tile = m_bridgeTiles.GetValue(i);

			if (AIBridge.IsBridgeTile(north_tile) && (AIBridge.GetOtherBridgeEnd(north_tile) == south_tile)) {
			    local old_bridge = AIBridge.GetBridgeID(north_tile);

                local bridge_list = AIBridgeList_Length(AIMap.DistanceManhattan(north_tile, south_tile) + 1);
                bridge_list.Valuate(AIBridge.GetMaxSpeed);
                bridge_list.Sort(AIList.SORT_BY_VALUE, AIList.SORT_DESCENDING);
				if (bridge_list.Count() > 0) {
			        local new_bridge = bridge_list.Begin();
			        if (TestBuildBridge().TryBuild(AIVehicle.VT_ROAD, new_bridge, north_tile, south_tile)) {
			            AILog.Info("Bridge at tiles " + north_tile + " and " + south_tile + " upgraded from " + AIBridge.GetName(old_bridge) + " to " + AIBridge.GetName(new_bridge));
					}
				}
			}
		}
	}

    function addVehicle(return_vehicle = false) {
        if(AIVehicleList_Depot(this.m_depotTile).Count() >= this.MAX_VEHICLE_COUNT) {
            return null;
        }

		if (m_engine != null && AIEngine.IsValidEngine(m_engine) && AIEngine.IsBuildable(m_engine)) {
            local vehicle = TestBuildRoadVehicle().TryBuild(this.m_depotTile, this.m_engine);
	        if(vehicle != null) {
    		    if (TestRefitRoadVehicle().TryRefit(vehicle, Utils.getCargoId(m_cargoClass))) {
	    		    AIOrder.AppendOrder(vehicle, m_depotTile, AIOrder.AIOF_SERVICE_IF_NEEDED | AIOrder.OF_NON_STOP_INTERMEDIATE);
                    AIOrder.AppendOrder(vehicle, m_stationFrom, AIOrder.AIOF_NON_STOP_INTERMEDIATE);
                    AIOrder.AppendOrder(vehicle, m_stationTo, AIOrder.AIOF_NON_STOP_INTERMEDIATE);

                    AIVehicle.StartStopVehicle(vehicle);

			        m_lastVehicleAdded = AIDate.GetCurrentDate();
             		m_sentToDepotList.RemoveItem(vehicle);
		            m_renewList.RemoveItem(vehicle);

                    if (return_vehicle) {
		                return vehicle;
		            } else {
		                return 1;
		            }
	    		}
		    	else {
			        AIVehicle.SellVehicle(vehicle);
			        return null;
			    }
            }
            else {
                return null;
            }
		}
		else {
		    return null;
		}
    }

    function addVehiclesToNewRoute(cargoClass) {
        if(AIVehicleList_Depot(this.m_depotTile).Count() >= this.MAX_VEHICLE_COUNT) {
            return null;
        }

        local stationDistance = AITile.GetDistanceManhattanToTile(m_stationFrom, m_stationTo);
        local buyVehicleCount = cargoClass == AICargo.CC_PASSENGERS ? START_VEHICLE_COUNT : MIN_VEHICLE_START_COUNT;

        buyVehicleCount += stationDistance / 20;
        if(buyVehicleCount > (MAX_VEHICLE_COUNT).tointeger()) {
            buyVehicleCount = (MAX_VEHICLE_COUNT).tointeger();
        }

        local swap_m_stationFrom = m_stationFrom;
		local swap_m_stationTo = m_stationTo;
        for(local i = 0; i < buyVehicleCount; ++i) {
            if(i % 2 == 1) {
                local swap = m_stationFrom;
                m_stationFrom = m_stationTo;
                m_stationTo = swap;
            }

            if(addVehicle()) {
                AILog.Info("Added " + AIEngine.GetName(this.m_engine) + " on new route from " + AITown.GetName(m_cityFrom)
                    + " to " + AITown.GetName(m_cityTo) + "!");
            }
        }
		if (AIVehicleList_Depot(this.m_depotTile).Count() == 0) {
		    m_lastVehicleAdded = 0;
		}

        //swap back
        m_stationFrom = swap_m_stationFrom;
        m_stationTo = swap_m_stationTo;
    }

    function sendVehicleToDepot(vehicle) {
        if ((!m_sentToDepotList.HasItem(vehicle)) && (!m_renewList.HasItem(vehicle)) && AIVehicle.GetState(vehicle) != AIVehicle.VS_CRASHED) {

            AIVehicle.SendVehicleToDepot(vehicle);

            m_lastVehicleRemoved = AIDate.GetCurrentDate();

            AIOrder.SetOrderFlags(vehicle, 0, AIOrder.AIOF_STOP_IN_DEPOT | AIOrder.AIOF_NON_STOP_INTERMEDIATE);

            AILog.Info(AIVehicle.GetName(vehicle) + " on route from " + AITown.GetName(m_cityFrom)
                + " to " + AITown.GetName(m_cityTo) + " has been sent to its depot!");

            return 1;
        }

        return 0;
    }

    function sendNegativeProfitVehiclesToDepot() {
        local vehicleList = AIVehicleList_Depot(m_depotTile);
        vehicleList.Valuate(AIVehicle.GetAge);
        vehicleList.KeepAboveValue(2 * 365);

        for(local vehicle = vehicleList.Begin(); vehicleList.HasNext(); vehicle = vehicleList.Next()) {
            AIController.Sleep(1);
            if((AIVehicle.GetProfitLastYear(vehicle)) < 0) {
                if(sendVehicleToDepot(vehicle)) {
                    m_sentToDepotList.AddItem(vehicle, 0);
                }
            }
        }
    }

    function sendLowProfitVehiclesToDepot(routeManager) {
        local roadVehicleList = AIVehicleList();
        roadVehicleList.Valuate(AIVehicle.GetVehicleType);
        roadVehicleList.KeepValue(AIVehicle.VT_ROAD);

        if(LuDiAIAfterFix.MAX_TOWN_VEHICLES - 50 > roadVehicleList.Count()) {
            return;
        }

        if((AIDate.GetCurrentDate() - m_lastVehicleRemoved) > 60) {
            local vehicleList = AIVehicleList_Depot(m_depotTile);
            vehicleList.Valuate(AIVehicle.GetAge);
            vehicleList.KeepAboveValue(2 * 365);


            for (local vehicle = vehicleList.Begin(); vehicleList.HasNext(); vehicle = vehicleList.Next()) {
                AIController.Sleep(1);
                local cargoId = Utils.getCargoId(m_cargoClass);
		        local station1 = AIStation.GetStationID(m_stationTo);
		        local station2 = AIStation.GetStationID(m_stationFrom);
		        local cargoWaiting1via2 = AICargo.GetDistributionType(cargoId) == AICargo.DT_MANUAL ? 0 : AIStation.GetCargoWaitingVia(station1, station2, cargoId);
		        local cargoWaiting1any = AIStation.GetCargoWaitingVia(station1, AIStation.STATION_INVALID, cargoId);
                local cargoWaiting1 = cargoWaiting1via2 + cargoWaiting1any;
		        local cargoWaiting2via1 = AICargo.GetDistributionType(cargoId) == AICargo.DT_MANUAL ? 0 : AIStation.GetCargoWaitingVia(station2, station1, cargoId);
		        local cargoWaiting2any = AIStation.GetCargoWaitingVia(station2, AIStation.STATION_INVALID, cargoId);
                local cargoWaiting2 = cargoWaiting2via1 + cargoWaiting2any;

                if ((cargoWaiting1 + cargoWaiting2 < 150) ||
                    (AIVehicle.GetProfitLastYear(vehicle) < (highestProfitLastYear(routeManager) / 6))) {
                    if(sendVehicleToDepot(vehicle)) {
                        m_sentToDepotList.AddItem(vehicle, 0);
                    }
                }
            }
        }
    }

    function sellVehiclesInDepot() {
        for(local vehicle = m_sentToDepotList.Begin(); m_sentToDepotList.HasNext(); vehicle = m_sentToDepotList.Next()) {
            if(AIVehicle.IsStoppedInDepot(vehicle)) {

                local vehicle_name = AIVehicle.GetName(vehicle);
    			m_sentToDepotList.RemoveItem(vehicle);
                AIVehicle.SellVehicle(vehicle);
				Utils.RepayLoan();

                AILog.Info(vehicle_name + " on route from " + AITown.GetName(m_cityFrom)
                    + " to " + AITown.GetName(m_cityTo) + " has been sold!");

            }
        }

        for(local vehicle = m_renewList.Begin(); m_renewList.HasNext(); vehicle = m_renewList.Next()) {
            if(AIVehicle.IsStoppedInDepot(vehicle)) {

                m_renewList.RemoveItem(vehicle);
                AIVehicle.SellVehicle(vehicle)
				Utils.RepayLoan();

				local renewed_vehicle = addVehicle(true);
                if(renewed_vehicle != null) {
                    AILog.Info(AIVehicle.GetName(renewed_vehicle) + " on route from " + AITown.GetName(m_cityFrom)
                        + " to " + AITown.GetName(m_cityTo) + " has been renewed!");

                }
            }
        }
    }

    function addVehicleToRoute() {
        if(!m_activeRoute || (AIDate.GetCurrentDate() -  m_lastVehicleAdded < 90)) {
            return null;
        }
		
		if (m_lastVehicleAdded == 0) {
		    return addVehiclesToNewRoute(m_cargoClass);
		}
        
        local cargoId = Utils.getCargoId(m_cargoClass);
		local station1 = AIStation.GetStationID(m_stationTo);
		local station2 = AIStation.GetStationID(m_stationFrom);
		local cargoWaiting1via2 = AICargo.GetDistributionType(cargoId) == AICargo.DT_MANUAL ? 0 : AIStation.GetCargoWaitingVia(station1, station2, cargoId);
		local cargoWaiting1any = AIStation.GetCargoWaitingVia(station1, AIStation.STATION_INVALID, cargoId);
        local cargoWaiting1 = cargoWaiting1via2 + cargoWaiting1any;
		local cargoWaiting2via1 = AICargo.GetDistributionType(cargoId) == AICargo.DT_MANUAL ? 0 : AIStation.GetCargoWaitingVia(station2, station1, cargoId);
		local cargoWaiting2any = AIStation.GetCargoWaitingVia(station2, AIStation.STATION_INVALID, cargoId);
        local cargoWaiting2 = cargoWaiting2via1 + cargoWaiting2any;

        if((cargoWaiting1 > 50 && cargoWaiting2 > 50) || (cargoWaiting1 + cargoWaiting2 > 300)) {
            if(addVehicle()) {
                AILog.Info("Added " + AIEngine.GetName(this.m_engine) + " on existing route from " + AITown.GetName(m_cityFrom)
                    + " to " + AITown.GetName(m_cityTo) + "!");
            }
        }
    }

    function renewVehicles() {
        local vehicleList = AIVehicleList_Depot(m_depotTile);
        vehicleList.Valuate(AIVehicle.GetAgeLeft);
        vehicleList.KeepBelowValue(365);

        for(local vehicle = vehicleList.Begin(); vehicleList.HasNext(); vehicle = vehicleList.Next()) {
            AIController.Sleep(1);
            if(sendVehicleToDepot(vehicle)) {
                m_renewList.AddItem(vehicle, 0);
            }
        }
    }

    function expandStations() {
	    local result = 0;
	    if (!m_activeRoute) return result;
		
		local vehicleList = AIVehicleList_Depot(m_depotTile);
        vehicleList.Valuate(AIVehicle.IsArticulated);
		vehicleList.KeepValue(1);
		local articulated = vehicleList.Count() != 0;
		
        local population = AITown.GetPopulation(m_cityFrom);
        if((population / 1000).tointeger() > m_expandedFromCount + 1) {
            if(BuildManager().buildTownStation(m_cityFrom, m_cargoClass, m_stationFrom, m_cityTo, articulated) != null) {
                ++m_expandedFromCount;
                result = 1;
            }
        }

        population = AITown.GetPopulation(m_cityTo);

        if((population / 1000).tointeger() > m_expandedToCount + 1) {
            local tile = BuildManager().buildTownStation(m_cityTo, m_cargoClass, m_stationTo, m_cityFrom, articulated);
            if(tile != null) {
                //AISign.BuildSign(tile, "exp");
                ++m_expandedToCount;
                result = 1;
            }
        }

        return result;
    }
	
	function removeIfUnserviced() {
	    local vehicleList = AIVehicleList_Depot(m_depotTile);
		local fully_removed = true;
		if (vehicleList.Count() == 0 && (AIDate.GetCurrentDate() -  m_lastVehicleAdded >= 90) && m_lastVehicleAdded != 0) {
		    m_activeRoute = false;
		
		    local fromTiles = AITileList_StationType(AIStation.GetStationID(m_stationFrom), m_cargoClass == AICargo.CC_PASSENGERS ? AIStation.STATION_BUS_STOP : AIStation.STATION_TRUCK_STOP);
			for (local tile = fromTiles.Begin(); fromTiles.HasNext(); tile = fromTiles.Next()) {
			    if (!TestRemoveRoadStation().TryRemove(tile)) {
				    m_stationFrom = AIBaseStation.GetLocation(AIStation.GetStationID(tile));
				    fully_removed = false;
				}
			}
			
			local toTiles = AITileList_StationType(AIStation.GetStationID(m_stationTo), m_cargoClass == AICargo.CC_PASSENGERS ? AIStation.STATION_BUS_STOP : AIStation.STATION_TRUCK_STOP);
			for (local tile = toTiles.Begin(); toTiles.HasNext(); tile = toTiles.Next()) {
			    if (!TestRemoveRoadStation().TryRemove(tile)) {
				    m_stationTo = AIBaseStation.GetLocation(AIStation.GetStationID(tile));
				    fully_removed = false;
				}
			}
			
			if (!TestRemoveRoadDepot().TryRemove(m_depotTile)) {
			    fully_removed = false;
			}
			
			if (fully_removed) {
			    AILog.Warning("Removed unserviced road route from " + AITown.GetName(m_cityFrom) + " to " + AITown.GetName(m_cityTo));
		        return true;
			}
		}
		return false;
	}

    function saveRoute() {
        local route = [];

        route.append(m_cityFrom);
        route.append(m_cityTo);
        route.append(m_stationFrom);
        route.append(m_stationTo);
        route.append(m_depotTile);

        local bridgeTilesTable = {};
		for (local bridge = m_bridgeTiles.Begin(), i = 0; m_bridgeTiles.HasNext(); bridge = m_bridgeTiles.Next(), ++i) {
		    bridgeTilesTable.rawset(i, [bridge, m_bridgeTiles.GetValue(bridge)]);
		}
		route.append(bridgeTilesTable);
		
        route.append(m_cargoClass);

        route.append(m_lastVehicleAdded);
        route.append(m_lastVehicleRemoved);
        route.append(m_activeRoute);
		route.append(m_expandedFromCount);
		route.append(m_expandedToCount);

        return route;
    }

    function loadRoute(data) {
        local cityFrom = data[0];
        local cityTo = data[1];
        local stationFrom = data[2];
        local stationTo = data[3];
        local depotTile = data[4];
		
		local bridgeTiles = AIList();
		local bridgeTable = data[5];
		local i = 0;
		while(bridgeTable.rawin(i)) {
		    local tile = bridgeTable.rawget(i);
			bridgeTiles.AddItem(tile[0], tile[1]);
			++i;
		}

        local cargoClass = data[6];
        local route = Route(cityFrom, cityTo, stationFrom, stationTo, depotTile, bridgeTiles, cargoClass, 1);

        route.m_lastVehicleAdded = data[7];
        route.m_lastVehicleRemoved = data[8];
        route.m_activeRoute = data[9];
		route.m_expandedFromCount = data[10];
		route.m_expandedToCount = data[11];

        return [route, i];
    }
}